(provide 'refactorerl-dupcode)

(require 'ediff)
(require 'ediff-util)
(require 'ediff-wind)

(require 'widget)
(require 'refactorerl-depanal)

(eval-when-compile
  (require 'wid-edit))


;; Dupcode search - similar to ri:search_duplicates/0
(defun refactorerl-search-dupcode ()
  ""
  (interactive)
	(search-and-show-dupcode "" "" "" "" "" ""))

(defun refactorerl-show-dupcode ()
  "Function to call refactorerl show dupcode function"
  (interactive)
  (setf code-to-show (car (read-from-string (read-from-minibuffer "Name of the duplicated code to show:"))))
  (refac-send/callbacks ('get_dupcode_result code-to-show)
  	 ('status "Duplicated code")
	 (:reply (ok result)
	 	(if (equal result "")
	 	 (print "No duplicated code with this name")
	     (process-dupcode-info result))))
  ) ;get_dupcode_result

(defvar filesvar "")
(defvar minlenvar "10")
(defvar minnumvar "2")
(defvar overlapvar "0")
(defvar filenamevar "")
(defvar namevar "")
;; Advanced dupcode search
(defun refactorerl-dupcode-advanced ()
  "Widget to set dupcode options"
  (interactive)
  (switch-to-buffer "*Duplicated code analysis*")
  (kill-all-local-variables)
  (make-local-variable 'filesvar)
  (make-local-variable 'minlenvar)
  (make-local-variable 'minnumvar)
  (make-local-variable 'overlapvar)
  (make-local-variable 'filenamevar)
  (make-local-variable 'namevar)          
  (let ((inhibit-read-only t))
    (erase-buffer))
  (let ((all (overlay-lists)))
    ;; Delete all the overlays.
    (mapcar 'delete-overlay (car all))
    (mapcar 'delete-overlay (cdr all)))
  (widget-insert " * Function and module dependencies *\n\n")
  (widget-insert "\nFiles: (separated with spaces)\n")
  (widget-create 'editable-field
	 :size 13
	 :value ""
	 :notify (lambda (widget &rest ignore) 
			   (setq filesvar (widget-value widget))
			   (widget-setup))
	 "")
  (widget-insert "\nMinimum length: \n")
  (widget-create 'editable-field
	 :size 13
	 :value "10"
	 :notify (lambda (widget &rest ignore) 
			   (setq minlenvar (widget-value widget))
			   (widget-setup))
	 "")
  (widget-insert "\nMinimum number: \n")
  (widget-create 'editable-field
	 :size 13
	 :value "2"
	 :notify (lambda (widget &rest ignore) 
			   (setq minnumvar (widget-value widget))
			   (widget-setup))
	 "")
  (widget-insert "\nOverlap: \n")
  (widget-create 'editable-field
	 :size 13
	 :value "0"
	 :notify (lambda (widget &rest ignore) 
			   (setq overlapvar (widget-value widget))
			   (widget-setup))
	 "")
  (widget-insert "\nOutput filename: \n")
  (widget-create 'editable-field
	 :size 13
	 :value ""
	 :notify (lambda (widget &rest ignore) 
			   (setq filenamevar (widget-value widget))
			   (widget-setup)))
  (widget-insert "\nName: \n")
  (widget-create 'editable-field
	 :size 13
	 :value ""
	 :notify (lambda (widget &rest ignore) 
			   (setq namevar (widget-value widget))
			   (widget-setup))
 	)
  (widget-create 'editable-field
		 :size 13
		 :value ""
		 :notify (lambda (widget &rest ignore) 
				   (setq gnodevar (widget-value widget))
				   (widget-setup))
		 "")
  (widget-insert "\n\n")
  (widget-insert "\n\n")
  (widget-create 'push-button
		 :notify (lambda (&rest ignore)
		 		(setf buffer (current-buffer))
		 		(kill-buffer buffer)
			    (search-and-show-dupcode filesvar minlenvar minnumvar overlapvar filenamevar namevar))
		 "Analise")
  (widget-insert " ")
  (widget-create 'push-button
		 :notify (lambda (&rest ignore) 
			   (kill-this-buffer))
		 "Cancel")
  (use-local-map widget-keymap)
  (widget-setup)
)

;; All parameters must be strings
(defun search-and-show-dupcode (files minlen minnum overlap filename name)
  "Calls RefactorErl's duplicated_code_initial_clone function with given options"
  (interactive)
  (let ((arglist (makearglist `(('files files ,(split-string files)) ('minlen minlen ,(string-to-number minlen))
								('minnum minnum ,(string-to-number minnum)) ('overlap overlap ,(string-to-number overlap))
								('output filename) ('name name)) )
		  ))
	(progn
  	(refac-send/callbacks ('duplicated_code_initial_clone arglist)
  	 ('status "Duplicated code")
	 (:reply (ok result)
	     (process-dupcode-info result)))
  ))
)



(defun process-dupcode-info (dinfo)
	"Processes duplicated code info send by RefactorErl, generates links for duplicates
	and shows the number of them and creates a buffer for this"
	(setf saved-as (car dinfo))
	(setf list0 (cdr dinfo))
	(if (equal list0 '("")) (setf list nil)
						 (setf list (car list0)))
	(switch-to-buffer "Duplicated code result")		
	(let ((inhibit-read-only t))
		   (erase-buffer)
		   (set (make-local-variable 'last-file-dir) ""))
	(widget-insert "\n *** Duplicated code information: *** \n\n")
	(let ((ln (length list))) (widget-insert (concatenate 'string (prin1-to-string ln) " duplicates found. \n\n")))
	(widget-insert (concatenate 'string "Saved as " (prin1-to-string saved-as) "\n\n"))
	(lexical-let ((i 1) (thisbuf (current-buffer))) nil
	(dolist (res list)
		(progn (print i)
	 	(lexical-let ((ii i) (rs res)) (widget-create 'link
				 :notify (lambda (&rest ignore)
				   (dupcode-window-open rs thisbuf))
		 (concatenate 'string (number-to-string i) ". duplicated code")))
		(widget-insert "\n")
		(setf i (+ i 1)))))
    (use-local-map widget-keymap)
	(widget-setup)
	(toggle-read-only 0)
	(delete-other-windows)
)


(defvar code-A nil)
(defvar code-B nil)

(defun dupcode-window-open (result all-duplicates-buf)
	(setq items 
			(loop for infopiece in result 
					
						collect (destructuring-bind ((fn filename) (st (stposl stposc)) (end (endposl endposc)) (nds nodes)) infopiece  (message (prin1-to-string filename))
	 			`(quote (item :tag ,filename :value ((fn ,filename) (st (,stposl ,stposc)) (end (,endposl ,endposc)))))))) ;(nds ,,nodes)
	 (message (prin1-to-string items))
	(progn  
	 	(delete-other-windows)
	 	(setq wind-ctrl (selected-window))
	 	(split-window-vertically)
	 	(other-window 1)
	 	(setq wind-A (selected-window))
	 	(split-window-vertically)
	 	(other-window 1)
	 	(setq wind-B (selected-window))
	  	(setf buf-ctrl (get-buffer-create "CtrlBuf"))
	 	(setf buf-A (get-buffer-create "BufA"))
	 	(setf buf-B (get-buffer-create "BufB"))
		(set-window-buffer wind-A buf-A)
		(set-window-buffer wind-B buf-B)
		(select-window wind-ctrl)
		(switch-to-buffer buf-ctrl)
		  ;(setf def_value `(nth 4 (nth 1 (car items)))) ;(nth 1 (nth 2 (nth 1 (car items))))) (message (prin1-to-string def_value))
		(widget-insert "Please select the codes you would like to compare below! \n \n")
		(eval (dupcode-file-chooser "Code A" 'code-A items))
		(eval (dupcode-file-chooser "Code B" 'code-B items))
  		(widget-create 'push-button
		 :notify (lambda (&rest ignore) 
					(if (and code-A code-B)
						(progn (dupcode-window-load-texts buf-A buf-B wind-A wind-B code-A code-B))
						(message "")
					))
		 "Ok")
		(widget-insert "\t")
		(lexical-let ((all-dup-buf all-duplicates-buf))
	  		(widget-create 'push-button
			 :notify (lambda (&rest ignore)
			 					(kill-buffer buf-ctrl)
			 					(delete-other-windows)
							 	(switch-to-buffer all-dup-buf))
			 "Back")
			(use-local-map widget-keymap))
			(widget-setup) 
		(fit-window-to-buffer wind-ctrl)
		(message (prin1-to-string wind-A)) (message (prin1-to-string wind-B)) (message (prin1-to-string wind-ctrl))
	)
)

(defun dupcode-file-chooser (label var items)
	(interactive)
	(lexical-let ((vr var) (items items) (val (nth 1 (car (nth 4 (nth 1 (car items))))))) ;(message (prin1-to-string val))
	(concatenate 'list
	`(widget-create 'menu-choice
		 :tag ,label
		 :value nil
		 :help-echo "Choose code duplicates!"
		 :notify '(lambda (widget &rest ignore)
		 			(message (prin1-to-string (widget-value widget)))
					(setf ,vr (widget-value widget)))
		 )
	 items
	))
)

(defun dupcode-window-load-texts (buf-A buf-B wind-A wind-B code-A code-B)
	(progn
		(select-window wind-A)
		(message (prin1-to-string (selected-window)))
		(destructuring-bind ((fn filename) (st (stposl stposc)) (end (endposl endposc))) code-A 
			(progn (message (prin1-to-string code-A))
			(refac-visit-in-current-buffer filename :start-line stposl :start-col stposc :end-line endposl :end-col endposc)) )
		(select-window wind-B)	
				(message (prin1-to-string (selected-window)))
		(destructuring-bind ((fn filename) (st (stposl stposc)) (end (endposl endposc))) code-B 
			(refac-visit-in-current-buffer filename :start-line stposl :start-col stposc :end-line endposl :end-col endposc))
	)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; UNUSED

(defun jump-to-dup (result)
	 (let ((inhibit-read-only t))
	       (erase-buffer)
	       (set (make-local-variable 'last-file-dir) ""))
	 (widget-insert "Duplicated code in the following files: \n")
	 (dolist (dupcode result)
	 	(progn
	 		(destructuring-bind ((fn filename) (st (stposl stposc)) (end (endposl endposc)) (nds nodes)) dupcode 
	 			(refac-widget-link filename :start-line stposl :start-col stposc :end-line endposl :end-col endposc))
	 		(widget-insert "\n")
	 	))
)

(defun ifnotextremal-int (int-elem ret)
  	"If a string is not -1 then it returns ret in a list (cause of append), 
  	else it returns nil"
  	(interactive)
  	(if (equal int-elem "-1")
  		nil
  		(list ret) )
)
