%%% -*- coding: latin-1 -*-

%%% The  contents of this  file are  subject to  the Erlang  Public License,
%%% Version  1.1, (the  "License");  you may  not  use this  file except  in
%%% compliance  with the License.  You should  have received  a copy  of the
%%% Erlang  Public License  along  with this  software.  If not,  it can  be
%%% retrieved at http://plc.inf.elte.hu/erlang/
%%%
%%% Software  distributed under  the License  is distributed  on an  "AS IS"
%%% basis, WITHOUT  WARRANTY OF ANY  KIND, either expressed or  implied. See
%%% the License  for the specific language governing  rights and limitations
%%% under the License.
%%%
%%% The Original Code is RefactorErl.
%%%
%%% The Initial Developer of the  Original Code is Eötvös Loránd University.
%%% Portions created  by Eötvös Loránd University are Copyright 2009,
%%% Eötvös Loránd University. All Rights Reserved.

%%% @author Tamas Hoffman <hoffmantamas@caesar.elte.hu>

%% -*- mode: nitrogen -*-

-module (investigations).
-compile(export_all).
-include_lib("referl_ui/include/nitrogen.hrl").
-include_lib("nitrogen/include/wf.hrl").
-include_lib("kernel/include/file.hrl").
-include("records.hrl").
-vsn("$Rev: 8854 $ ").

-define(Query(Label,Query),#link{class=mylink,
                                text=Label, 
                                postback={predefquery,Query,Node}},#br{}).
                                
-record(wdata, {label,startnode,container,name,hidden=false,text}).

%%% ============================================================================
%%% HTML output

main() -> 
    wf:session(selectednode_data, undefined),
    wf:session(current_window, undefined),
    wf:session(query_startwindow, undefined),
    wf:session(current_query, undefined),
    wf:session(investigation,[]),
    wf:session(nodetext,""),
    wf:session(inv_name,"Unnamed"),
    nitrogen_lib:main("investigations.html").

title() -> "RefactorErl Queries".

other() ->
    HiddenData=#hidden{id=hiddendata,text=""},
    HiddenButton=#link{id=hiddenbutton,body="",postback=node_clicked},
    HiddenButton2=#link{id=load_file_button,postback=load_investigation},
    #table{class=width_100,rows=[#tablerow{cells=[
        #tablecell{body=HiddenData},
        #tablecell{body=HiddenButton},
        #tablecell{body=HiddenButton2}]}]}.

menu() ->
    wf:wire(save_button, 
            savename, 
            #validate { validators=[#is_required { text="Required." }]}),
    StartLabel=#label{text="Starting function (m:f/a):"},
    StartButton=#table{rows=#tablerow{cells=[
        #tablecell{body=#textbox_autocomplete{id=start_textbox,
                                              tag=funs,
                                              style="width:150px"}},
        #tablecell{body=#button{id=start_button,
                        text="Start inv.",
                        postback=start_investigation}}]}},
    SaveButton=#table{rows=#tablerow{cells=[
        #tablecell{body=#textbox{id=savename,style="width:150px"}},
        #tablecell{body=#button{id=save_button,text="Save",
        postback=save_investigation}}]}},
    Rename=#table{rows=#tablerow{cells=[
        #tablecell{body=#label{id=invname,text="Unnamed"}},
        #tablecell{body="&nbsp;"},
        #tablecell{body=#button{id=rename_inv,
                                text="Rename",
                                postback=rename_inv}}]}},
    BrowserLabel=#label{text="Investigations:"},
    Browser=#panel{id=file_browser,
                   body="<div id='file_browser'></div>"},
    Result1=#panel{id=executed_query_placeholder,body=""},
    Result2=#panel{id=query_result_placeholder,body=""},
    #table{class=width_100,rows=[
        #tablerow{cells=#tablecell{body=Rename}},
        #tablerow{cells=#tablecell{body=StartLabel}},
        #tablerow{cells=#tablecell{body=StartButton}},
        #tablerow{cells=#tablecell{body=SaveButton}},
        #tablerow{cells=#tablecell{body=BrowserLabel}},
        #tablerow{cells=#tablecell{body=Browser}},
        #tablerow{cells=#tablecell{body=Result1}},
        #tablerow{cells=#tablecell{body=Result2}}
    ]}.

investigations() -> 
    #panel{id=mainwrapper,body=#span{id=contextmenu,text="",class=contextmenu}}.
    
get_startnode() ->
    wf:set(savename,""),
    Start=wf:q(start),
    case Start of
        undefined -> ok;
        _ ->
            [D1,D2]=string:tokens(Start, "|"),
            TT=list_to_existing_atom(D1),
            Id=list_to_integer(D2),
            Clause={'$gn',TT,Id},
            add_window(Clause,0,Clause,"")
    end,
    [].
    
logout()->
    nitrogen_lib:logout_link(wf:user()).

%%% ============================================================================
%%% Handlers for postback events

event(logout)->
    nitrogen_lib:logout(wf:user());

event(rename_inv) ->
    wf:replace(invname,#textbox{id=invname,postback=rename_inv_final}),
    wf:wire(rename_inv,#hide{});
    
event(rename_inv_final) ->
    NewName=wf:q(invname),
    wf:session(inv_name,NewName),
    wf:replace(invname,#label{id=invname,text=NewName}),
    wf:wire(rename_inv,#show{});

event(start_investigation) ->
    wf:set(savename,""),
    reset_windows(),
    StartFunction=wf:q(start_textbox),
    Result=re:split(StartFunction,"([:/])",[{return,list},group]),
    case Result of
        [[Module,":"],[Function,"/"],[Arity]] when length(Module)>0, 
                length(Function)>0, length(Arity)>0 ->
            Fun=try
                    ?NITRO_SERVICES:get_function(Module,Function,Arity)
                catch
                    _:_ -> []
                end,
            if
                Fun==[] -> wf:wire(#alert{text="Function doesn't exist!"});
                true -> 
                    Clauses=?NITRO_SERVICES:get_clauses(Fun),
                    wf:session(current_window,0),
                    continue({continue, jump}, 
                        {Clauses,"Select starting clause",none,hd(Fun)})
            end;
        _ -> wf:wire(#alert{text="Bad function format!"})
    end;

event({docustomquery_context,Node})->
    Query=wf:q(context_query_box),
    Type=?NITRO_SERVICES:get_type(Node),
    event({docustomquery,Node,Query,Type});

event({predefquery,Query,Node})->
    event({docustomquery,Node,Query,none});

event({docustomquery,Node,Query,Type})->
    NotExprQuery=lists:sublist(Query,5)/="@expr",
    if
        (Query/=[]) andalso (hd(Query)/=$@) -> 
            wf:wire(#alert{text="No global queries allowed!"});
        Query=="" ->
            ok;
        NotExprQuery ->
            PNode=?NITRO_SERVICES:get_proper_node(Node),
            execute_query(Query,{PNode,Node},Type);
        true -> 
            execute_query(Query,{Node,Node},Type)
    end;

event({query_jump_to_node,Node,Num})->
    Inv=wf:session(investigation),
    ButtonId="'query_result"++integer_to_list(Num)++"'",
    wf:wire(#script{script="
        try{
            if(lastcoloredresult!=null) 
                $(lastcoloredresult).removeClass('selected_queryresult')
            $(obj("++ButtonId++")).addClass('selected_queryresult');
            lastcoloredresult=obj("++ButtonId++");
        }catch(ex){}"}),
    WindowNum=tree_get_lastid(Inv),
    WindowId="wi"++integer_to_list(WindowNum),
    try
        CNode=?NITRO_SERVICES:get_container(Node),
        WindowBody=try
            ?NITRO_SERVICES:generate_node(CNode,WindowNum)
        catch
            request_denied -> "Database currently in use"
        end,
        WindowData=tree_get_node_data(Inv,WindowNum),
        wf:session(investigation,
            tree_modify_node(Inv,WindowNum,WindowData#wdata{container=CNode,
                                                            text=WindowBody})),
        wf:replace(WindowId,#panel{id=WindowId,body=WindowBody}),
        StartLine=integer_to_list(
            ?NITRO_SERVICES:get_node_linenum_comment(CNode)),
        wf:wire(#script{script=
        "generatelinenums(obj('"++WindowId++"'),
        obj('linenums"++integer_to_list(WindowNum)++"'),"++StartLine++");"}),
        show_result(Node,WindowNum)
    catch
        request_denied -> wf:wire(#alert{text="Database currently in use!"})
    end;

event(node_clicked)->
    [D1,D2,Window]=string:tokens(wf:q(hiddendata), "|"),
    TT=list_to_existing_atom(D1),
    Id=list_to_integer(D2),
    Node={'$gn',TT,Id},
    Type=?NITRO_SERVICES:get_type(Node),
    CurrentSel=wf:session(selectednode_data),
    CurrentWin=wf:session(current_window),
    if
    (CurrentSel==Node) and (Window==CurrentWin) -> 
        cleanup();
    true ->
        wf:wire(#script{script="
        try{
            elem=selectednode;
            if(lastcolored!=null)
            {
                $(lastcolored).removeClass('selectednode');
                var children=lastcolored.childNodes;
                for(var i=0;i<children.length;i++) 
                    $(children[i]).removeClass('selectednode');
            }
            $(elem).addClass('selectednode');
            var children=elem.childNodes;
            for(var i=0;i<children.length;i++) 
                $(children[i]).addClass('selectednode');
            lastcolored=elem;
            $.post('ajax_handler', { nodetext: $(elem).text()} );
        }catch(ex){}"}),
        Rows=main:get_predef_queries(Node,Type,invs),
        wf:replace(contextmenu,
            #panel{id=contextmenu,body=Rows,class=contextmenu}),
        wf:wire(#script{script="positionContextMenu(selectednode,false);"}),
        wf:session(current_window,list_to_integer(Window)),
        wf:session(selectednode_data,Node)
    end;
    
event(save_investigation) ->
    FileName=wf:q(savename),
    event({save_investigation,FileName,fun() -> ok end});
    
event({save_investigation,FileName,EndFun}) ->
    delete_empty_window(),
    BaseDir=mnesia:system_info(directory)++"/investigations",
    filelib:ensure_dir(BaseDir++"/"++FileName++".inv"),
    {ok, File} = file:open(BaseDir++"/"++FileName++".inv", [write]),
    Fun=fun(Tree) ->
            io:format(File, "~p.", [{wf:session(inv_name),
                                     ?NITRO_HELPER:get_database_hash(),Tree}]),
            file:close(File),
            EndFun()
        end,
    put_positions_in_tree(wf:session(investigation),Fun);

event(load_investigation) ->
    File=wf:session(selected_file),
    case lists:suffix("/", File) of
        true -> ok;
        false -> event({load_investigation,File})
    end;

event({load_investigation,File}) ->
    wf:set(savename,filename:basename(File,".inv")),
    CurrentHash=?NITRO_HELPER:get_database_hash(),
    FileExists=filelib:is_file(File),
    {Name,Hash,Investigations}=if
        FileExists==true -> 
            {Result, LoadedData} = file:consult(File),
            if
                (Result==ok) and (LoadedData/=[]) -> hd(LoadedData);
                true -> {"Error loading investigation!",CurrentHash,[]}
            end;
        true ->
            {"Error loading investigation!",CurrentHash,[]}
    end,
    if
        Hash/=CurrentHash -> wf:wire(#alert{text=
            "Database has changed since last visit, "++
            "investigation might not be usable."});
        true -> ok
    end,
    reset_windows(),
    wf:replace(invname,#label{id=invname,text=Name}),
    put_windows(Investigations);
    
event({window_options,Window}) ->
    cleanup(),
    Inv=wf:session(investigation),
    WindowId="wi"++integer_to_list(Window),
    wf:replace(contextmenu,#panel{id=contextmenu,body=[],class=contextmenu}),
    {_,_,Data,Children}=tree_get_node(Inv,Window),
    HaveKeepButton=(Window==tree_get_lastid(Inv)) 
                and (wf:session(current_query)/=undefined) 
                and (wf:session(current_window)/=1)
                and (Data#wdata.container/=empty),
    if
        HaveKeepButton ->
            wf:insert_bottom(contextmenu, 
                #button{text="New window",postback={window_keep,Window}}),
            wf:insert_bottom(contextmenu,#br{});
        true ->
            ok
    end,
    
    HaveDeleteButton=(Children==[]) 
                and (Data#wdata.container/=empty),
    if
        HaveDeleteButton ->
            wf:insert_bottom(contextmenu, 
                #button{text="Delete window",postback={window_delete,Window}}),
            wf:insert_bottom(contextmenu,#br{});
        true ->
            ok
    end,
    if
        (Data#wdata.hidden) and (Data#wdata.hidden/=deny)->
            wf:insert_bottom(contextmenu, 
                #button{text="Show contents",postback={window_show,Window}});
        (not Data#wdata.hidden) and (Data#wdata.hidden/=deny) ->
            wf:insert_bottom(contextmenu, 
                #button{text="Hide contents",postback={window_hide,Window}});
        true -> ok
    end,
    wf:insert_bottom(contextmenu,#br{}),
    wf:insert_bottom(contextmenu, 
        #button{text="Move to new investigation",
                postback={window_moveto_new,Window}}),
    wf:insert_bottom(contextmenu,#br{}),
    wf:insert_bottom(contextmenu, 
        #button{text="Rename",
                postback={window_rename,Window}}),
    wf:wire(#script{script="positionContextMenu(obj('"++WindowId++"'),true);"});

event({window_delete,Window}) ->
    wf:wire(#script{script="$(obj('contextmenu')).css('display','none');"}),
    delete_window(Window);
    
event({window_rename,Window}) ->
    LabelName="wn"++integer_to_list(Window),
    wf:wire(#script{script="$(obj('contextmenu')).css('display','none');"}),
    wf:replace(LabelName,#textbox{id=LabelName,
                                  postback={window_rename_final,Window}});
                                  
event({window_rename_final,Window}) ->
    LabelName="wn"++integer_to_list(Window),
    NewName=wf:q(LabelName),
    wf:wire(#script{script="$(obj('contextmenu')).css('display','none');"}),
    Inv=wf:session(investigation),
    WindowData=tree_get_node_data(Inv,Window),
    wf:session(investigation,
        tree_modify_node(Inv,Window,WindowData#wdata{name=NewName})),
    wf:replace(LabelName,#label{id=LabelName,text=NewName});                                

event({window_moveto_new,Window}) ->
    LabelName="wn"++integer_to_list(Window),
    wf:wire(#script{script="$(obj('contextmenu')).css('display','none');"}),
    wf:replace(LabelName,[#label{id=tmplab,text="Save new investigation as:"},
                          #textbox{id=LabelName,
                                   postback={window_moveto_new_final,Window}}]);

event({window_moveto_new_final,Window}) ->
    LabelName="wn"++integer_to_list(Window),
    NewName=getnametext(LabelName,"unnamed_new"),
    wf:replace(tmplab,#label{id=tmplab,text="Save old investigation as:"}),
    wf:replace(LabelName,#textbox{id=LabelName,
                          postback={window_moveto_new_final2,Window,NewName}});

event({window_moveto_new_final2,Window,NewName}) ->
    LabelName="wn"++integer_to_list(Window),
    OldName=getnametext(LabelName,"unnamed_old"),
    event({moveto_new_investigation,Window,OldName,NewName});

event({moveto_new_investigation,Window,OldName,NewName}) ->
    NewTree=tree_get_subtree(wf:session(investigation),Window),
    RemainingTree=tree_delete_subtree(wf:session(investigation),Window),
    WindowData=tree_get_node_data(RemainingTree,Window),
    NewFile=mnesia:system_info(directory)++"/investigations/"++NewName++".inv",
    wf:session(investigation,
        tree_modify_node(RemainingTree,Window,
            WindowData#wdata{container={investigation,NewName}})),
    event({save_investigation,OldName,fun() -> 
        wf:session(investigation,NewTree),
        event({save_investigation,NewName,fun() -> 
                                          event({load_investigation,NewFile}) 
                                          end}) 
                                    end});

event({window_hide,Window}) ->
    Inv=wf:session(investigation),
    WindowText=integer_to_list(Window),
    WindowId="wi"++WindowText,
    wf:wire(#script{script="$(obj('contextmenu')).css('display','none');"}),
    #wdata{container=CN}=tree_get_node_data(Inv,Window),
    case CN of 
        {_,_,_} ->
    WindowBody=?NITRO_SERVICES:node_to_text(CN),
    wf:replace("hb"++WindowText,
        #button{id="hb"++WindowText,text="Show",postback={window_show,Window}}),
    wf:replace(WindowId,#panel{id=WindowId,body=WindowBody}),
    StartLine=integer_to_list(
        ?NITRO_SERVICES:get_node_linenum_comment(CN)),
    wf:wire(#script{script=
        "generatelinenums(obj('wi"++integer_to_list(Window)++"'),
            obj('linenums"++integer_to_list(Window)++"'),"++StartLine++");"}),
    WindowData=tree_get_node_data(Inv,Window),
    wf:session(investigation,
        tree_modify_node(Inv,Window,WindowData#wdata{hidden=true})),
    wf:wire(#script{script="
        $(obj('w"++WindowText++"')).width('auto');
        $(obj('w"++WindowText++"')).height('auto');
        $(obj('wd"++WindowText++"')).width('auto');
        $(obj('wd"++WindowText++"')).height('auto');
        repaintEverything();"});
        _ -> ok
    end;

event({window_show,Window}) ->
    Inv=wf:session(investigation),
    WindowText=integer_to_list(Window),
    wf:wire(#script{script="$(obj('contextmenu')).css('display','none');"}),
    WindowId="wi"++WindowText,
    {_,_,WindowData,Children}=tree_get_node(Inv,Window),
    case WindowData#wdata.container of 
        {_,_,_} ->
    WindowBody=try
        ?NITRO_SERVICES:generate_node(WindowData#wdata.container,Window)
    catch
        request_denied -> "Request to database was denied!"
    end,
    StartLine=integer_to_list(
        ?NITRO_SERVICES:get_node_linenum_comment(WindowData#wdata.container)),
    wf:replace("hb"++WindowText,
        #button{id="hb"++WindowText,text="Hide",postback={window_hide,Window}}),
    wf:replace(WindowId,#panel{id=WindowId,body=WindowBody}),
    wf:wire(#script{script=
        "generatelinenums(obj('wi"++integer_to_list(Window)++"'),
            obj('linenums"++integer_to_list(Window)++"'),"++StartLine++");"}),
    wf:session(investigation,
        tree_modify_node(Inv,Window,WindowData#wdata{hidden=false})),
    [rehighlight_nodes(Inv,W,Window) || W <- Children],
    wf:wire(#script{script="repaintEverything();"});
        _ -> ok
    end;
    
event({window_keep,_}) ->
    wf:wire(#script{script="$(obj('contextmenu')).css('display','none');"}),
    cleanup(),
    {StartNode,Query,Text}=wf:session(current_query),
    add_window(empty,wf:session(query_startwindow),StartNode,
               Text++"<br/>"++Query).

%%% ============================================================================
%%% Autocomplete

autocomplete_enter_event(SearchTerm, funs) ->
    Funs=case ?NITRO_SERVICES:execute_system_query("mods.funs", 
                                                   {query_display_opt,
                                                    [],
                                                    needed_pattern,$:}) of
        {result, no_result} -> [];
        {result, ResultFuns} -> ResultFuns
    end,    
    DataF=lists:map(fun(Elem)-> 
                           {struct, [{id, list_to_binary(Elem)}, 
                                     {label, list_to_binary(Elem)} , 
                                     {value, list_to_binary(Elem) }]} 
                    end,Funs),
    List = [{struct,[{id, Id }, {label, Label}, {value, Value}]} || 
            {struct,[{id, Id }, {label, Label}, {value, Value}]} <- DataF, 
            string:str(string:to_lower(binary_to_list(Label)), 
                       string:to_lower(SearchTerm)) > 0],
    mochijson2:encode(List);

autocomplete_enter_event(SearchTerm, _Tag) ->
    PosQueries=?NITRO_SERVICES:do_autocomplete(SearchTerm),
    Data=lists:map(fun({Label,Value})-> 
                          {struct, [{id, list_to_binary(Value)}, 
                                    {label, list_to_binary(Label)} , 
                                    {value, list_to_binary(Value) }]} 
                   end,PosQueries),
    mochijson2:encode(Data).

autocomplete_select_event({struct,[{<<"id">>,_},{<<"value">>, Value}]},_Tag) ->
    wf:flash(Value),
    ok.

%%% ============================================================================
%%% Handlers for continous events 

continue({continue, put_positions},{{NodeId,Parent,Data,List},NS,Fun,Sum})->
    [X,Y,W,H]=string:tokens(wf:q(hiddendata), "|"),
    NewSum=Sum++[{NodeId,Parent,Data,X,Y,W,H,List}],
    put_positions_in_tree_(NS,Fun,NewSum);

continue({continue, jump}, {error,Query,Error}) ->
    main:query_result_handler({Query,0,0},{{result, no_result},
                              {warning,no_warning},{error,Error}});
continue({continue, jump}, {FilteredNodes,Query,_,StartNode}) ->
    FileList=lists:keysort(2,lists:map(
                fun(X) -> {X, ?NITRO_SERVICES:get_file(X)} end,
                FilteredNodes)),
    GroupedList=main:group_files(FileList),
    Result=lists:map(fun(X) ->
        InnerList=lists:map(fun(Y) ->
                    {{node,?NITRO_SERVICES:get_node_linenum(Y),Y},
                    ?NITRO_SERVICES:node_to_text(Y)}
                end,
                tl(X)),
        {group_by,{nopos,lists:last(string:tokens (hd(X), "/"))},list,InnerList} 
                     end, GroupedList),
    wf:session(query_startwindow,wf:session(current_window)),
    wf:session(current_query,{StartNode,Query,wf:session(nodetext)}),
    main:query_result_handler({Query,0,0},{{result, Result},
                              {warning,no_warning},{error,no_error}}),
    if
        FilteredNodes==[] -> ok;
        true ->
            CNode=?NITRO_SERVICES:get_container(hd(FilteredNodes)),
            add_window(CNode,wf:session(query_startwindow),StartNode,
                       wf:session(nodetext)++"<br/>"++Query),
            wf:wire(#script{script="
                try{$(obj('query_result1')).click();}catch(ex){}"})
    end.

%%% ============================================================================
%%% Helper functions

%% @doc Clean things associated with a selected node
cleanup() ->
    wf:wire(#script{script="
        try{
            if(lastcolored!=null)
            {
                $(lastcolored).removeClass('selectednode');
                var children=lastcolored.childNodes;
                for(var i=0;i<children.length;i++) 
                    $(children[i]).removeClass('selectednode');
            }
            lastcolored=null;
            $(obj('contextmenu')).css('display','none');
        }catch(ex){}"}),
    wf:session(selectednode_data, undefined).

%% @doc Show result node in given window
show_result(Node,Window) ->
    {_,NodeT,NodeId}=Node,
    NodeTypeText=atom_to_list(NodeT),
    NodeIdText=integer_to_list(NodeId),
    wf:wire(#script{script="
    try{
        var elem;
        var window=obj('wi"++integer_to_list(Window)++"');
        var elems=document.getElementsByName('jump_"
            ++NodeTypeText++NodeIdText++"');
        for(var i=0;i<elems.length;i++) if(jQuery.contains(window,elems[i])) 
            {elem=elems[i];break;}
        if(elem!=null)
        {
            showcontext=false;
            scrolltonode(elem,window);
            if(elem.tagName=='A')
            {
                $(elem).click();
            }
            else
            {
                $(elem).addClass('selectednode');
                lastcolored=elem;
                selectednode=elem;
                cd.value='"++NodeTypeText++"|"++NodeIdText
                           ++"|"++integer_to_list(Window)++"';
                $(cb).click();
            }
        }
    }catch(ex){}"}).

%% @doc Execute given query from given node
execute_query(Query,{PNode,Node},Type) ->
    delete_empty_window(),
    Fun=main:check_query_start(Query,{PNode,Node},Type),
    cleanup(),
    wf:replace(executed_query_placeholder,
               #panel{id=executed_query_placeholder,
                      body=[#p{body=?NITRO_CORE:whitespaces_to_html(Query),
                               class="query_title"},
                            #br{},
                            #p{body="Running query, please wait.."},
                            #br{}]}),
    wf:continue({continue, jump}, Fun, 60*60*1000).

%%% ============================================================================
%%% Window management

%% @doc Add a new investigaton window with ContainerNode's contents.
%% If Parent is not 0, connect it to given parent window's StartNode, 
%% with given label
add_window(CNode,Parent,{_,Type,NId}=StartNode,Label) ->
    NextId=tree_get_lastid(wf:session(investigation))+1,
    IdText=integer_to_list(NextId),
    NodeId="jump_"++atom_to_list(Type)++integer_to_list(NId),
    ParentId="w"++integer_to_list(Parent),
    WindowId="w"++integer_to_list(NextId),
    DivId="wd"++integer_to_list(NextId),
    {_,_,_,List}=tree_get_node(wf:session(investigation),Parent),
    Children=integer_to_list(length(List)),
    WindowText=case CNode of 
        {_,_,_} -> 
            try
                ?NITRO_SERVICES:generate_node(CNode,NextId)
            catch
                request_denied -> "Request to database was denied!"
            end;
        {investigation,FileName} -> 
            [FileName++".inv",
             #br{},
             #link{text="View this investigation",
                   postback={load_investigation,
                       mnesia:system_info(directory)++"/investigations/"
                       ++FileName++".inv"}}];
        _ -> "&nbsp;"
    end,
    NewDiv=#panel{id=WindowId,class=inv_window,body=#table{rows=[
        #tablerow{class=drg,cells=[
            #tablecell{class=width_15,body=#button{text="Options",
                                    postback={window_options,NextId}}},
            #tablecell{class=width_15,body=#button{id="hb"++IdText,text="Hide",
                                                postback={window_hide,NextId}}},
            #tablecell{body=#label{id="wn"++IdText,
                                   text="Window"++IdText}}]},
        #tablerow{cells=[
            #tablecell{colspan=2,body=#panel{
                       id=DivId,
                       class=inv_window_inner,
                       body=#table{class=[drg,all_100],rows=#tablerow{cells=[
                            #tablecell{body=#panel{id="linenums"++IdText,
                                                   body=[],
                                                   class=inv_lnums}},
                            #tablecell{body=#panel{id="wi"++IdText,
                                                   body=WindowText}}]}}}}]}]}},
    wf:insert_bottom(mainwrapper, NewDiv),
    case CNode of 
        {_,_,_} ->
            StartLine=integer_to_list(
                ?NITRO_SERVICES:get_node_linenum_comment(CNode)),
            wf:wire(#script{script=
            "generatelinenums(obj('wi"++IdText++"'),
            obj('linenums"++IdText++"'),"++StartLine++");"});
        _ -> ok
    end,
    if
        Parent>0 ->
            wf:wire(#script{script="
                $(obj('"++WindowId++"')).css('left',
                    parseInt($(obj('"++ParentId++"')).css('left'),10)
                    +$(obj('"++ParentId++"')).width()+50);
                $(obj('"++WindowId++"')).css('top',
                    parseInt($(obj('"++ParentId++"')).css('top'),10)+
                    +$(obj('"++ParentId++"')).height()*"++Children++");
                jsPlumb.connect({source:obj('"++ParentId++"'),
                                 target:obj('"++WindowId++"'),
                anchors:[ calculateRelativePosition(obj('"++ParentId++"'),
                    '"++NodeId++"'),[0,0.5,-1,0] ],
                overlays:[ [ 'Label', { label:'"++Label++"', 
                    location:0.6, cssClass:'inv_label' } ]]});"});
        true -> wf:wire(#script{script="
                $(obj('"++WindowId++"')).css('top',30);"})   
    end,
    
    wf:wire(#script{script="
        setDragResize(obj('"++WindowId++"'),obj('"++DivId++"'));"}),
    
    WindowData=#wdata{label=Label,startnode=StartNode,container=CNode,
                      text=WindowText,name="Window"++integer_to_list(NextId)},
    wf:session(investigation,tree_add_node(wf:session(investigation),
        Parent,WindowData)).
        
add_window_custom(CNode,Parent,{_,Type,NId}=StartNode,Label,
                  Id,X,Y,W,H,Text,Hidden,Name) ->
    IdText=integer_to_list(Id),
    NodeId="jump_"++atom_to_list(Type)++integer_to_list(NId),
    ParentId="w"++integer_to_list(Parent),
    WindowId="w"++IdText,
    DivId="wd"++IdText,
    {CanHide,WindowText}=case CNode of 
        {_,_,_} -> 
            try
                TypeTest=?NITRO_SERVICES:get_type(CNode),
                if
                    TypeTest==error -> 
                        {deny,Text};
                    Hidden -> 
                        {Hidden,?NITRO_SERVICES:node_to_text(CNode)};
                    not Hidden -> 
                        {Hidden,?NITRO_SERVICES:generate_node(CNode,Id)}
                end
            catch
                _ -> {deny,Text}
            end;
        {investigation,FileName} -> 
            {Hidden,
            [FileName++".inv",
             #br{},
             #link{text="View this investigation",
                   postback={load_investigation,
                       mnesia:system_info(directory)++"/investigations/"
                       ++FileName++".inv"}}]};
        _ -> {Hidden,"&nbsp;"}
    end,
    HideButton=if
        Hidden -> #tablecell{class=width_15,
                             body=#button{id="hb"++IdText,
                                          text="Show",
                                          postback={window_show,Id}}};
        true -> #tablecell{class=width_15,
                           body=#button{id="hb"++IdText,
                                        text="Hide",
                                        postback={window_hide,Id}}}
    end,
  NewDiv=#panel{id=WindowId,class=inv_window,body=#table{rows=[
        #tablerow{class=drg,cells=[
            #tablecell{class=width_15,
                     body=#button{text="Options",postback={window_options,Id}}},
                     HideButton,
            #tablecell{body=#label{id="wn"++IdText,text=Name}}]},
        #tablerow{cells=[
            #tablecell{colspan=3,body=#panel{
                       id=DivId,
                       class=inv_window_inner,
                       body=#table{class=[drg,all_100],rows=#tablerow{cells=[
                            #tablecell{body=#panel{id="linenums"++IdText,
                                                   body=[],
                                                   class=inv_lnums}},
                            #tablecell{body=#panel{id="wi"++IdText,
                                                   body=WindowText}}]}}}}]}]}},
    wf:insert_bottom(mainwrapper, NewDiv),
   case CNode of 
        {_,_,_} ->
            StartLine=integer_to_list(
                ?NITRO_SERVICES:get_node_linenum_comment(CNode)),
            wf:wire(#script{script=
            "generatelinenums(obj('wi"++IdText++"'),
            obj('linenums"++IdText++"'),"++StartLine++");"});
        _ -> ok
    end,
    wf:wire(#script{script="
        $(obj('"++WindowId++"')).width("++W++");
        $(obj('"++WindowId++"')).height("++H++");
        $(obj('"++DivId++"')).width("++W++"-20);
        $(obj('"++DivId++"')).height("++H++"-24);
        $(obj('"++DivId++"')).css('max-height',"++H++");
        $(obj('"++WindowId++"')).css('left',"++X++");
        $(obj('"++WindowId++"')).css('top',"++Y++");"}),
    if
        Parent>0 ->
            wf:wire(#script{script="
                jsPlumb.connect({source:obj('"++ParentId++"'),
                                 target:obj('"++WindowId++"'),
                anchors:[ calculateRelativePosition(obj('"++ParentId++"'),
                    '"++NodeId++"'),[0,0.5,-1,0] ],
                overlays:[ [ 'Label', { label:'"++Label++"', 
                    location:0.5, cssClass:'inv_label' } ]]});"});
        true -> ok       
    end,
    
    wf:wire(#script{script="
        setDragResize(obj('"++WindowId++"'),obj('"++DivId++"'));"}),
    
    WindowData=#wdata{label=Label,startnode=StartNode,container=CNode,
                      text=Text,hidden=CanHide,name=Name},
    wf:session(investigation,tree_add_node(wf:session(investigation),
        Parent,WindowData)).

%% @doc Delete window that has given id
delete_window(Window) ->
    cleanup(),
    WindowId="w"++integer_to_list(Window),
    {_,Parent,Data,_}=tree_get_node(wf:session(investigation),Window),
    wf:session(investigation,
               tree_delete_node(wf:session(investigation),Window)),
    if
        Parent/=0 ->
            {_,_,_,Children}=tree_get_node(wf:session(investigation),Parent),
            StartNodes=[(tree_get_node_data(
                        wf:session(investigation),W))#wdata.startnode 
                            || W <- Children],
            RemoveHighLight=not lists:member(Data#wdata.startnode,StartNodes),
            ParentId="wi"++integer_to_list(Parent),
            {_,Type,NId}=Data#wdata.startnode,
            NodeId="jump_"++atom_to_list(Type)++integer_to_list(NId),
            if
                RemoveHighLight ->
                    wf:wire(#script{script="
                    nodes=document.getElementsByName('"++NodeId++"');
                    for(var i=0;i<nodes.length;i++) 
                        if(jQuery.contains(obj('"++ParentId++"'),nodes[i])) 
                        {
                            node=nodes[i];
                            break;
                        }
                    $(node).removeClass('inv_startnode');"});
                true -> ok
            end,
            wf:wire(#script{script=
                "jsPlumb.detachAllConnections($(obj('"++WindowId++"')));"});
    true -> ok
    end,
    wf:remove(WindowId).

%% @doc Remove all windows and connections
reset_windows() ->
    wf:session(investigation,[]),
    wf:session(current_window, undefined),
    wf:session(query_startwindow, undefined),
    wf:session(current_query, undefined),
    cleanup(),
    wf:replace(executed_query_placeholder,
               #panel{id=executed_query_placeholder}),
    wf:replace(query_result_placeholder,
               #panel{id=query_result_placeholder}),
    wf:replace(mainwrapper,#panel{id=mainwrapper,
        body=#span{id=contextmenu,text="",class=contextmenu}}),
    wf:wire(#script{script="jsPlumb.detachEveryConnection();czoom=1;zoom=1;"}).

%% @doc Create windows from given list
put_windows([]) -> ok;
put_windows([{Id,Parent,Data,X,Y,W,H,_}|NS]) ->
    add_window_custom(Data#wdata.container,Parent,Data#wdata.startnode,
                     Data#wdata.label,Id,X,Y,W,H,Data#wdata.text,
                     Data#wdata.hidden,Data#wdata.name),
    put_windows(NS).
    

%%% ============================================================================
%%% Functions for tree data structure 

%% @doc Add node with given id and data to tree
tree_add_node(Tree,Id,Data) -> 
    tree_add_node_(Tree,Id,tree_get_lastid(Tree)+1,Data).
tree_add_node_([],_,_,Data) -> [{1,0,Data,[]}];
tree_add_node_([{NodeId,Parent,NodeData,NodeList}|NS],Id,NewId,Data) 
when Id==NodeId ->
    NewNodeList=NodeList++[NewId],
    [{NodeId,Parent,NodeData,NewNodeList}|NS++[{NewId,NodeId,Data,[]}]];
tree_add_node_([N|NS],Id,NewId,Data) -> [N|tree_add_node_(NS,Id,NewId,Data)].

%% @doc Get node with given id from tree
tree_get_node([],_) -> {0,0,notfound,[]};
tree_get_node([{NodeId,_,_,_}=N|_],Id) when Id==NodeId -> N;
tree_get_node([_|NS],Id) -> tree_get_node(NS,Id).

%% @doc Get data of node with given id from tree
tree_get_node_data([],_) -> #wdata{container=notfound};
tree_get_node_data([{NodeId,_,Data,_}|_],Id) when Id==NodeId -> Data;
tree_get_node_data([_|NS],Id) -> tree_get_node_data(NS,Id).

%% @doc Get highest id number in tree
tree_get_lastid(Tree) -> tree_get_lastid_(Tree,0).
tree_get_lastid_([],Max) -> Max;
tree_get_lastid_([{Id,_,_,_}|NS],Max) when Id>Max -> tree_get_lastid_(NS,Id);
tree_get_lastid_([_|NS],Max)-> tree_get_lastid_(NS,Max).

%% @doc Delete node with given id from tree, removing all connections as well
tree_delete_node(Tree,Id) ->
    {_,Parent,_,_}=tree_get_node(Tree,Id),
    tree_delete_node_(Tree,Id,Parent).
tree_delete_node_([],_,_) -> [];
tree_delete_node_([{NodeId,_,_,_}|NS],Id,Parent) when Id==NodeId -> 
    tree_delete_node_(NS,Id,Parent);
tree_delete_node_([{NodeId,NParent,NodeData,NodeList}|NS],Id,Parent) 
when Parent==NodeId -> 
    NewList=lists:delete(Id,NodeList),
    [{NodeId,NParent,NodeData,NewList}|tree_delete_node_(NS,Id,Parent)];
tree_delete_node_([N|NS],Id,Parent) -> [N|tree_delete_node_(NS,Id,Parent)].

%% @doc Modify data in node with given id in tree
tree_modify_node([],_,_) -> [];
tree_modify_node([{NodeId,Parent,_,List}|NS],Id,NewData) when Id==NodeId -> 
    NewNode={NodeId,Parent,NewData,List},
    [NewNode|tree_modify_node(NS,Id,NewData)];
tree_modify_node([N|NS],Id,NewData) -> [N|tree_modify_node(NS,Id,NewData)].

%% @doc Get subtree starting from given node in tree
tree_get_subtree(Tree,Id) -> 
    {_,_,Data,NodeList}=tree_get_node(Tree,Id),
    [{Id,0,Data,NodeList}|tree_get_subtree_(Tree,NodeList,[])].
tree_get_subtree_([],_,NewTree) -> NewTree;
tree_get_subtree_([{NodeId,_,_,List}=N|NS],Ids,NewTree) ->
    IsMember=lists:member(NodeId,Ids),
    if
        IsMember -> tree_get_subtree_(NS,Ids++List,NewTree++[N]);
        true -> tree_get_subtree_(NS,Ids,NewTree)
    end.
    
%% @doc Delete subtree starting from given node in tree excluding that node
tree_delete_subtree(Tree,Id) ->
    {_,_,_,NodeList}=tree_get_node(Tree,Id),
    tree_delete_subtree_(Tree,NodeList,Tree).
tree_delete_subtree_(_,[],NewTree) -> NewTree;
tree_delete_subtree_([{NodeId,_,_,List}|NS],Ids,NewTree) ->
    IsMember=lists:member(NodeId,Ids),
    if
        IsMember -> tree_delete_subtree_(NS,lists:delete(NodeId,Ids)
                    ++List,tree_delete_node(NewTree,NodeId));
        true -> tree_delete_subtree_(NS,Ids,NewTree)
    end.

%%% ============================================================================
%%% Helper functions

%% @doc Delete the empty window if there is one
delete_empty_window() ->
    LastId=tree_get_lastid(wf:session(investigation)),
    #wdata{container=CN}=tree_get_node_data(wf:session(investigation),LastId),
    if
        CN==empty ->
            delete_window(LastId);
        true ->
            ok
    end.

%% @doc Put window position information in nodes
put_positions_in_tree(Tree,Fun) ->
    put_positions_in_tree_(Tree,Fun,[]).
put_positions_in_tree_([],Fun,Sum) -> Fun(Sum);
put_positions_in_tree_([{NodeId,_,_,_}=N|NS],Fun,Sum) ->
    ParentId="w"++integer_to_list(NodeId),
    wf:wire(#script{script="
        x=parseInt($(obj('"++ParentId++"')).css('left'),10);
        y=parseInt($(obj('"++ParentId++"')).css('top'),10);
        w=$(obj('"++ParentId++"')).width();
        h=$(obj('"++ParentId++"')).height();
        obj('hiddendata').value=x+'|'+y+'|'+w+'|'+h;"}),
    wf:continue({continue,put_positions},fun()->{N,NS,Fun,Sum} end, 60*60*1000).

%% @doc Highlight W's startnode in StartW
rehighlight_nodes(Inv,W,StartW) ->
    #wdata{startnode={_,T,Id}}=tree_get_node_data(Inv,W),
    wf:wire(#script{script="
        var elem=null;
        var window=obj('wi"++integer_to_list(StartW)++"');
        var elems=document.getElementsByName('jump_"++
            atom_to_list(T)++integer_to_list(Id)++"');
        for(var i=0;i<elems.length;i++) if(jQuery.contains(window,elems[i])) 
            {elem=elems[i];break;}
        $(elem).addClass('inv_startnode');"}).

%% @doc Get text from textbox, if empty set it to default
getnametext(Id,Default) ->
    Text=wf:q(Id),
    if
        Text==undefined orelse Text==[] -> Default;
        true -> Text
    end.
